@@ -11,7 +11,7 @@ from typing import Any, Dict, List, Optional, Set, Tuple
 import pandas as pd
 from mitosheet.code_chunks.code_chunk import CodeChunk
 from mitosheet.code_chunks.step_performers.merge_code_chunk import MergeCodeChunk
-from mitosheet.errors import (make_incompatible_merge_headers_error,
+from mitosheet.errors import (get_recent_traceback, make_incompatible_merge_headers_error,
                               make_incompatible_merge_key_error)
 from mitosheet.state import DATAFRAME_SOURCE_MERGED, State
 from mitosheet.step_performers.step_performer import StepPerformer
@@ -31,28 +31,23 @@ class MergeStepPerformer(StepPerformer):
 
     @classmethod
     def step_version(cls) -> int:
-        return 3
+        return 4
 
     @classmethod
     def step_type(cls) -> str:
         return 'merge'
 
-    @classmethod
-    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
-        return params
-
     @classmethod
     def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
-        how = get_param(params, 'how')
-        sheet_index_one = get_param(params, 'sheet_index_one')
-        merge_key_column_id_one = get_param(params, 'merge_key_column_id_one')
-        selected_column_ids_one = get_param(params, 'selected_column_ids_one')
-        sheet_index_two = get_param(params, 'sheet_index_two')
-        merge_key_column_id_two = get_param(params, 'merge_key_column_id_two')
-        selected_column_ids_two = get_param(params, 'selected_column_ids_two')
+        how: str = get_param(params, 'how')
+        sheet_index_one: int = get_param(params, 'sheet_index_one')
+        sheet_index_two: int = get_param(params, 'sheet_index_two')
+        merge_key_column_ids: List[List[ColumnID]] = get_param(params, 'merge_key_column_ids')
+        selected_column_ids_one: List[ColumnID] = get_param(params, 'selected_column_ids_one')
+        selected_column_ids_two: List[ColumnID] = get_param(params, 'selected_column_ids_two')
 
-        merge_key_one = prev_state.column_ids.get_column_header_by_id(sheet_index_one, merge_key_column_id_one)
-        merge_key_two = prev_state.column_ids.get_column_header_by_id(sheet_index_two, merge_key_column_id_two)
+        merge_keys_one = prev_state.column_ids.get_column_headers_by_ids(sheet_index_one, list(map(lambda x: x[0], merge_key_column_ids)))
+        merge_keys_two = prev_state.column_ids.get_column_headers_by_ids(sheet_index_two, list(map(lambda x: x[1], merge_key_column_ids)))
 
         selected_columns_one = prev_state.column_ids.get_column_headers_by_ids(sheet_index_one, selected_column_ids_one)
         selected_columns_two = prev_state.column_ids.get_column_headers_by_ids(sheet_index_two, selected_column_ids_two)
@@ -66,10 +61,10 @@ class MergeStepPerformer(StepPerformer):
             prev_state.df_names,
             how,
             sheet_index_one,
-            merge_key_one,
+            merge_keys_one,
             selected_columns_one,
             sheet_index_two,
-            merge_key_two,
+            merge_keys_two,
             selected_columns_two
         )
         pandas_processing_time = perf_counter() - pandas_start_time
@@ -94,16 +89,7 @@ class MergeStepPerformer(StepPerformer):
         ]
     
     @classmethod
-    def get_modified_dataframe_indexes( # type: ignore
-        cls, 
-        how: str,
-        sheet_index_one: int,
-        merge_key_column_id_one: ColumnID,
-        selected_column_ids_one: List[ColumnID],
-        sheet_index_two: int,
-        merge_key_column_id_two: ColumnID,
-        selected_column_ids_two: List[ColumnID],
-    ) -> Set[int]:
+    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
         return {-1}
 
 def _execute_merge(
@@ -111,10 +97,10 @@ def _execute_merge(
         df_names: List[str],
         how: str,
         sheet_index_one: int,
-        merge_key_one: ColumnHeader, 
+        merge_keys_one: List[ColumnHeader], 
         selected_columns_one: List[ColumnHeader],
         sheet_index_two: int,
-        merge_key_two: ColumnHeader,
+        merge_keys_two: List[ColumnHeader],
         selected_columns_two: List[ColumnHeader]
     ) -> pd.DataFrame:
     """
@@ -127,9 +113,13 @@ def _execute_merge(
     if dfs[sheet_index_one].columns.nlevels != dfs[sheet_index_two].columns.nlevels:
         raise make_incompatible_merge_headers_error(error_modal=False)
 
-    if how == 'lookup':
+    # If there's no merge keys, we return an empty dataframe
+    if len(merge_keys_one) == 0 and len(merge_keys_two) == 0:
+        return pd.DataFrame()
+
+    if how == LOOKUP:
         # We drop duplicates to avoid pairwise duplication on the merge.
-        temp_df = dfs[sheet_index_two].drop_duplicates(subset=merge_key_two)
+        temp_df = dfs[sheet_index_two].drop_duplicates(subset=merge_keys_two)
         # We overwrite the how variable to 'left' so it can be used in the merge
         how_to_use = 'left'
     else:
@@ -153,10 +143,32 @@ def _execute_merge(
 
     try:
         if how == UNIQUE_IN_LEFT:
-            return df_one_cleaned.copy(deep=True)[~df_one_cleaned[merge_key_one].isin(df_two_cleaned[merge_key_two])]
+            df_two_cleaned = df_two_cleaned.drop_duplicates(subset=merge_keys_two) # Remove duplicates so lookup merge only returns first match
+            # Create a boolean index array for values only in the left sheet by turning on the indicator=True and then filtering on the result
+            bool_index_array = df_one_cleaned.merge(df_two_cleaned, left_on=merge_keys_one, right_on=merge_keys_two, how='left', indicator=True)['_merge'] == 'left_only'
+            # Get the final result by filtering on the boolean index array and then selecting the columns we want to keep
+            return df_one_cleaned.copy(deep=True)[bool_index_array][selected_columns_one].reset_index(drop=True)
         if how == UNIQUE_IN_RIGHT:
-            return df_two_cleaned.copy(deep=True)[~df_two_cleaned[merge_key_two].isin(df_one_cleaned[merge_key_one])]
+            df_one_cleaned = df_one_cleaned.drop_duplicates(subset=merge_keys_one) # Remove duplicates so lookup merge only returns first match
+            bool_index_array = df_one_cleaned.merge(df_two_cleaned, left_on=merge_keys_one, right_on=merge_keys_two, how='right', indicator=True)['_merge'] == 'right_only'
+            return df_two_cleaned.copy(deep=True)[bool_index_array][selected_columns_two].reset_index(drop=True)
         else:
-            return df_one_cleaned.merge(df_two_cleaned, left_on=[merge_key_one], right_on=[merge_key_two], how=how_to_use, suffixes=[f'_{suffix_one}', f'_{suffix_two}'])
+            return df_one_cleaned.merge(df_two_cleaned, left_on=merge_keys_one, right_on=merge_keys_two, how=how_to_use, suffixes=[f'_{suffix_one}', f'_{suffix_two}'])
     except ValueError:
+        # If we get a value error from merging two incompatible columns, we go through and check 
+        # to see which of the columns this is, so our error can be maximally informative
+        for merge_key_one, merge_key_two in zip(merge_keys_one, merge_keys_two):
+            merge_key_one_dtype = str(dfs[sheet_index_one][merge_key_one].dtype)
+            merge_key_two_dtype = str(dfs[sheet_index_two][merge_key_two].dtype)
+
+            if merge_key_one_dtype != merge_key_two_dtype:
+                raise make_incompatible_merge_key_error(
+                    merge_key_one=merge_key_one, 
+                    merge_key_one_dtype=merge_key_one_dtype,
+                    merge_key_two=merge_key_two, 
+                    merge_key_two_dtype=merge_key_two_dtype,
+                    error_modal=False
+                )
+
         raise make_incompatible_merge_key_error(error_modal=False)
+
